#define PROGNAME "display"
#include <stdio.h>
#include <graph.h>
#include "c:sys.types.h"
#include "c:sys.times.h"

struct tms start_time,stop_time;
long write_time;
#define HERTZ 50
#define TIME_OFFSET 2

#define XFS_2 0 /* top half of framestore coords */
#define YFS_2 512
#define TOP_HALF 1
#define BOTTOM_HALF 2

#define MAXX 512
#define MAXY 512
#define MAXLARGE 8              /* maximum enlargement */
#define DW 16                   /* dither matrix width */
#define DP 4                    /* dither "pixel" width */

#define MAPF "map.dat"          /* default mapfile */

#define MAPPLANES 8             /* planes of display */
#define FSTOREX 128             /* bytes */
#define FSTOREW 64              /* words */
#define FSTOREL 32              /* longwords */
#define MAPLENGTH 256           /* number of addresses in colour map */
#define BRIGHT 255              /* top of colour map */
#define IOSCALE  3              /* MAPPLANES - (no of planes of grey) */

#define GREYMASK 0377           /* masks for enable */
#define REDMASK 017
#define GREENMASK 0360


/* convert v for monochrome colour map, so as to drive all 3 channels */
#define grey(v) (v+(v<<5)+(v<<10))
#define redgreen(r,g)  ((r<<1)+(g<<6))

short header=1;
short buf[MAXX];
short lastbuf[MAXX];
short col = WHITE;
short thresh = 0;
short xsize=MAXX;
short xs=MAXX;
short frameno = 0;
short ysize=0;
short large=1;                    /* picture enlargement factor */
char  * topleft;                  /* position of topleft of image, offset from frame */
FILE *fp = NULL;
char stereo=0;                    /* stereo pairs as anaglyphs */

char hidden=1;                    /* display built in top half of frame store */
short disp_time=10;               /* display time */
short delay_time=0;               /* time in seconds of min delay */
#define FIDDLE FSTOREX
char clearflag=1;                 /* clear tophalf of fs unless -f or -o flag */
char clear_botflag=0;             /* to clear bottom half of frame store */

char sflag = 0;                   /* sole picture, displayed in the middle */
char dflag = 1;                   /* default is dither */
char overlay = 0;                 /* overlay one image on last */
char gflag =0;                    /* display thresholded gradients */
char zflag=0;                     /* dont use 32 greyval colourmap */
char *mapfile ;                   /* name of colour map file */

short dither [DW][DW];
short ditherp[DP][DP] = {   0, 16, 48, 96,
                           32, 64,112,160,
                           80,128,176,208,
                           142,192,224,240};

short lookup[2*MAPLENGTH];

main(argc,argv)
char **argv;
{
  short mask;  /* mask for enable */

  times(&start_time); /* time from clock */
  offset(0,0);
  argv++;

  while((argc>0)&&(*argv[0] == '-'))
  {
    switch((*argv)[1]) {
    case 'x':
      sscanf(*argv+2,"%hd",&xsize);
      header=0;
      break;
    case 'f':  /* frame number */ 
      sscanf(*argv+2,"%hd",&frameno);
      clearflag=0;
      break;
    case 'o':            /* overlay used only with -z */
      sscanf(*argv+2,"%hd",&col);
      overlay=1;
      zflag=1;    
      clearflag=0;
      break;
    case 'C': /*force clear bottom of frame store*/
      clear_botflag=1;
      break;
    case 's':
      sflag=1;
      break;
    case 'g':
      gflag=1;
      sscanf(*argv+2,"%hd",&thresh);
      break;
    case 'd':      /* no dither - use with retouch */
      dflag=0;
      break;
    case 'c':
      sscanf(*argv+2,"%hd",&col);
      break;
    case 'm':      /* mapfile name */
      mapfile= *argv+2;
      if((fp = fopen(mapfile,"r")) == NULL)
        if((fp = fopen(MAPF,"r")) == NULL)
          error("can't open mapfile");
      break;
    case 'z':  /* for simulated grey without colour map */
      zflag=1;
      break;
    case 'l': 
      sscanf(*argv+2,"%hd",&large);
      if (large<1)
        large=1;
      break;
    case 'a':   /* anaglyphs */
      stereo=1;
      break;
    case 'u':   /* dont use top half of frame store to hide pic */
      hidden=0;
      break;
    case 't':  /* set minimum running time for screen to change to show */
               /* top half of frame store */
      sscanf(*argv+2,"%hd",&delay_time);
      break;
    case 'T':  /* set maximum time for picture on top half of fs to be shown*/  
      sscanf(*argv+2,"%hd",&disp_time);
      break;
    default:
      error("bad option");
      break;
    }
    argc--;
    argv++;
  }

  makedither();  /* set up dither matrix */

  set_lookup();  /* default lookup table */
  readmap();     /* lookup table from file, if any */

  if(clearflag) clear_fs(TOP_HALF);

  if (stereo)
  {
    mask=GREENMASK;
    anamap();      /* set up colour map for red/green */
  }
  else
  {
    mask=GREYMASK;
  }

  if (argc<2)
    error("picture input on stdin not implemented");

  while (--argc>0) 
  {
    if((fp=fopen(*(argv++),"r"))==NULL)
      error("can't open picture file");
    if (stereo)
      display(mask,MAPPLANES/2);
    else if (dflag)
      display(mask,IOSCALE);
    else
      display(mask,0);
    fclose(fp);
   
    if (stereo)
    {
      if (mask==GREENMASK)
      {
        mask = REDMASK;
      }
      else
      {
        mask=GREENMASK;
        frameno++;
      }
    }
    else
      frameno++;
  }

  if(hidden){
    times(&stop_time);
    write_time = (stop_time.tms_utime - start_time.tms_utime)/HERTZ+TIME_OFFSET;
    wait(delay_time-write_time);
    setupdmap();   /* default colour map */
    if (zflag)        resetmap();  /* reset to level 1 graphics config */
                                   /* this causes the screen to BLINK */
    offset(XFS_2,YFS_2);
    wait(disp_time);
    if(clear_botflag) clear_fs(BOTTOM_HALF);
    offset(0,0);
  }

}


error(s)
char *s;
{
  fprintf(stderr,"%s: %s\n",PROGNAME,s);
  offset(0,0);
  exit(0);
}


makedither()
{
  register short i,j;
  short scale;

  for (i=0;i<DW;i++)
  {
    for (j=0;j<DW;j++)
    {
      dither[i][j] = ditherp[(i%DP)][(j%DP)] ;
      dither[i][j] +=  ((ditherp[(i/DP)][(j/DP)])/(DP*DP));
    }
  }

/* grey level display uses dither values >>(MAPPLANES-IOSCALE) */
/* anaglyph uses >> (MAPPLANES/2) */

  if ((!zflag)||stereo)
  {
    scale= stereo?MAPPLANES/2:MAPPLANES-IOSCALE;
    for (i=0; i<DW; i++)
      for (j=0; j<DW; j++)
        dither[i][j] = dither[i][j] >> scale;
  }
}


copybuf()
{
register short *lp, *p, i;

  for (p=buf,lp=lastbuf,i=0; i<xs; i++)
    *p++ = *lp++;
}


grad()
{
register short j,d;
register short *lp, *p,hv;

  p=buf;
  lp=lastbuf;

  if(ysize==0)
  {
    for(j=1;j<xs;j++)
    {
      d= *(p+1) - *p;
      *lp++ = *p;
      *p++ = ((d>thresh)||(d< -thresh))?BRIGHT:0;
    }
    *lp= *p;
  }
  else
  {
    for(j=1;j<xs;j++)
    {
      hv=0;
      d= *lp - *p;
      if ((d>thresh)||(d< -thresh))
        hv=BRIGHT;
      d= *(p+1) - *p;
      if ((d>thresh)||(d< -thresh))
        hv=BRIGHT;
      *lp++ = *p;
      *p++ = hv;
    }
      d= *lp- *p;
      *lp= *p;
      *p = ((d>thresh)||(d< -thresh))?BRIGHT:0;
  }
}



readline()
{
  register short *p, i;

  for (p=buf,i=0; i++ <xsize;)
    *p++ =(getc(fp))&0377;
}



enlarge(large)
{
  register short *pi, *pj, i;

  for (pi=buf+xsize,pj=buf+xs;pi>buf;)
  {
    --pi;
    for (i=0;i<large;i++)
      *--pj= *pi;
  }
}


border(x,y,size)
{
  enable(GREYMASK);
  if (zflag)
    colour(col);
  else
    colour(BRIGHT);

  line(x,y,x+size,y);
  line(x,y-size,x+size,y-size);
  line(x,y,x,y-size);
  line(x+size,y,x+size,y-size);
}




clear_fs(fs_half)
char fs_half;
{
register long  *p;
register short i;

  colour(BLACK);
  enable(GREYMASK);

  if((fs_half&01)==TOP_HALF) p = ((long *) frame + YFS_2*FSTOREL);
  else                       p =  (long *) frame;

  for (i=0; i<2048; i++)
  {
    *p++ = 0xFFFFFFFF;
    *p++ = 0xFFFFFFFF;
    *p++ = 0xFFFFFFFF;
    *p++ = 0xFFFFFFFF;
    *p++ = 0xFFFFFFFF;
    *p++ = 0xFFFFFFFF;
    *p++ = 0xFFFFFFFF;
    *p++ = 0xFFFFFFFF;
  }
}

set_lookup()  /* set up default lookup */
{
short i,c;

  for (i=0; i<MAPLENGTH; i++)
  {
    lookup[i] = i;
  }
  c=i;
  for (i=0;i<MAPLENGTH;i++)   
  {
    lookup[i+MAPLENGTH] = c;   /* saturate */
  }
}


setupdmap()  /* set up hardware colour map */
{
short i,c,cmax,cscale;

  cmax=(MAPLENGTH>>IOSCALE)-1;
  cscale = 0;
  if (!dflag)
    cscale=IOSCALE;

  for (i=0; i<MAPLENGTH; i++)
  {
    c=i>>cscale;
    c=(c>cmax)?cmax:c;
    colourmap[2*i+1] = grey(c);
  }
}

readmap()
{
  short i,v;
  if(fp == NULL)
  {
    return(0);
  }
  for (i=0;i<MAPLENGTH;i++)
  {
    v = 0377&getc(fp);
    lookup[i]=v;
  }
  fclose(fp);
}




anamap()  /* set up hardware colour map for anaglyph */
{
short i,r,g;

  for (i=0; i<MAPLENGTH; i++)
  {
    r=i&REDMASK;
    g=(i&GREENMASK)>>(MAPPLANES>>1);
    colourmap[2*i+1] = redgreen(r,g);
  }
}

disgrey_line(po,rp)
char *po;
short rp;   /* no of grey planes to be simulated by dither */
{
register unsigned short bit;
register short *pi,*pw,i;
register short *di, c, r, rmask;
short bp;
short *dirow;
char right;


/*display the line*/
  bp = MAPPLANES-rp;
  rmask = 1<<rp -1;

  if (zflag)
  {
    dirow = dither[(ysize&(DW-1))];
    for (pi=buf,pw= (short *)po,i=0; i<xs;pw++)
      for(di=dirow,bit=0100000;(bit!=0)&&(i++<xs);bit=(bit>>1))
      {
        if (lookup[(*pi++)] > *di++)
          colour(col);
        else
          colour(BLACK);
        *pw = bit;
      }
  }
  else if (stereo)    /* grey stereo */
  {
    dirow = dither[(ysize&(DW-1))];
    for (pi=buf,pw= (short *)po,i=0; i<xs;pw++)
    {
      for(di=dirow,bit=0100000;(bit!=0)&&(i++<xs);bit=(bit>>1))
      {
        c=lookup[(*pi++)];
        r = c & rmask;
        c = c>>rp;
        if (r > *di++)
          c++;
        c = (c>rmask)?rmask:c;
        c |= (c<<rp);
        colour(c);
        *pw = bit;
      }
    }
  }
  else
  {
    dirow = dither[(ysize&(DW-1))];
    for (pi=buf,pw= (short *)po,i=0; i<xs;pw++)
    {
      for(di=dirow,bit=0100000;(bit!=0)&&(i++<xs);bit=(bit>>1))
      {
        c=lookup[(*pi++)];
        r = c & rmask;
        c = c>>rp;
        if (r > *di++)
          c++;
        colour(c);
        *pw = bit;
      }
    }
  }

  return(0);
}



display(mask,rp)
short mask;   /* mask for enable - used to select red/green */
short rp;   /* no of grey planes to be simulated by dither */
{
char *po;
short xoff, yoff, i, maxframe;
long offs;

  if (header)
    xsize=getc(fp)&0377;


  xs=xsize*large;
  if(xs>MAXX)
    error("picture too large");
  if(large>MAXLARGE)
    error("enlargement factor too big");

  maxframe = YPIXELS/ (xs);
  if (sflag) {
    yoff = (YPIXELS>>1) + (xs>>1)/*+4*/;
    xoff = (XPIXELS>>4) - (xs>>4) +1;
  }
  else {
    if (frameno>=maxframe*maxframe)
      frameno=0;
    yoff = frameno/maxframe;
    xoff= frameno - yoff*maxframe;
    yoff = (yoff+1) *(xs)/*+4*/;
    xoff = xoff * ((xs>>3) + 4);
  }

  offs = (( FSTOREX*yoff +xoff) & (~01));    /* word aligned*/
  topleft = frame + offs;
  if(hidden) topleft=topleft+(((long)YFS_2)*FSTOREX)-FIDDLE;

  if (zflag)
  {
    if (overlay)
      enable(col);
    else
      enable(ALL);
  }
  else
    enable(mask);

  po = topleft;

  for (ysize=0; ysize<xs;)
  {
    readline();
    if (large>1)
      enlarge(large);
    for (i=0;i<large;i++,ysize++)
    {
      if (gflag)
        grad();
      disgrey_line(po,rp);
      po -= FSTOREX;
      if (gflag)
        copybuf();  /* reinstate buf */
    }
  }

  /* border((offs%FSTOREX)*8+1,offs/FSTOREX-1,xs-1);*/
}

#define CYCLE 82000L             /* no of delay loops per second */

wait(time)
long time;       /* in seconds */
{
  long i,max;
  
  if(time>0) {
    max=time*CYCLE;
    for(i=0;i<max;i++);
  }
}
