!  File  NMOUSE:MOUSE.INC

! Mouse system include file

%option "-low-nons"

%constinteger queue size = 24
%constinteger buffer size = 2048
%constinteger message header size = queuesize+24
%constinteger max message size = buffersize-messageheadersize

%constinteger queue pool size = 256
%constinteger buffer pool size = 128

%recordformat queue fm -
  (%byte lock %or %integer tag, %record(*)%name queue, forward, backward)

%recordformat semaphore fm (%record(queue fm)header,%record(*)%name owner,
                            %integer counter)
%recordformat mailbox fm   (%record(queue fm)header,%record(*)%name owner,
                            %record(semaphore fm)%name semaphore)
%recordformat run queue fm (%record(queue fm)header,%record(*)%name owner,
                            %bytename interrupt register)

%recordformat kernel object fm (%record(queue fm)header,
                                %record(*)%name owner,*,
                                %string(255-queuesize)name)

%recordformat message fm   (%record(queue fm)header,%record(*)%name owner,
                            %record(mailbox fm)%name reply,
   ({message} %integer size %or-
    {pipe}    %record(semaphore fm)%name space, info,
    {pipe}    %integer put,get,minspace,mininfo),
    {both}    %bytearray data(0:max message size-1))

%recordformat process fm   (%record(queue fm)header,%record(*)%name owner,
                            %record(run queue fm)%name run queue,
                            %integer queue quota, buffer quota,
   %integer vbr, ssp, cputime, starttime,
            priority, time slice size, time slice left,
            asn {address space number},
   %record(*)%name pti {page table index},
   %record(process fm)%name parent,youngest,elder,younger,
   %record(queue fm)%name grave,
   %string(63)name %or-
   %bytearray b(1:buffer size))
  {The rest of the first kilobyte contains the supervisor stack, on which the
  {register context is saved, its position being recorded in SSP field.  The
  {second kilobyte contains the "process-own" (i.e. module-global) data area.

%recordformat interrupt handler fm -
  (%record(queue fm)header %or %integer pc)

@16_11c0  %record(process fm)%name current process,
   {11c4} %record(queue fm)%name queue pool,
   {11c8} %record(mailbox fm)%name buffer pool,
   {11cc} %record(mailbox fm)%name name server mailbox,
   {11d0} %record(kernel object fm)%arrayname ko(1:queue pool size),
   {11d4} %record(message fm)%arrayname bu(1:buffer pool size),
   {11d8} %integer system elapsed time,
   {11dc} %integer idle time,
   {11e0} %record(run queue fm)%namearray run queue(0:7 {prio 0 is lowest}),
   {1200} %record(interrupt handler fm)%array int chain(1:7),
   {1270} %integer global code base,
   {1274} %integer global code limit,
   {1278} %integer interrupted priority,
   {127c} %record(process fm)%name rivalprocess,
   {1280} %integer ownbase,
   {1284} %integer stacklimit
   {1288}

@16_3f00 %integer cputype
@16_3ff0 %integer freebot,freetop,membot,memtop

! Loader/Imp specific stuff

%recordformat fe02 object fm -
  (%short flags,extra,%integer type,offset,%string(255)name)
%recordformat fe02 header fm -
  (%half tyver,flags,export,import,%integer codesize,%half reset,main,
   %integer ownsize,stack,%half dlim,chlim,spare1,spare2)
%recordformat fe02 indir fm -
  (%integer address %or -
   %half op1, %integer opd1, %half op2, %integer opd2)
%recordformat mar fm {combined file connection and module activation record} -
  (%record(mar fm)%name next,
   (%integer start %or %record(fe02 header fm)%name header),
   %integer size,gla,
   %string(255)name)

%recordformatspec fcb fm
%recordformat trap fm(%record(trap fm)%name next,%halfname mask,%integer gla,*)
%recordformat poa fm(%bytearray b(1:1024) %or-
  (%integer evlink %orrecord(trap fm)%name evtrap),
  %integerarray display(1:7),
  %byte event,sub,%short line,%integer extra,%string(255-32)message,
  %integer eventpc,%integerarray eventdisplay(1:7),eventregs(0:15),
  %integer scratch1,scratch2,*{excepth},%record(marfm)%name filelist,
  %integer heapbase,stacklimit,
  %record(fcb fm)%name curin,curout,
  %integer instream,outstream,
  %record(fcb fm)%namearray in,out(0:7),
  %string(255)cliparam)
  
%register(a5)%record(poa fm)%name poa

! Procedures provided in EPROM

@16_408 %integerfn romreadfile(%string(255)file,%integer where) {booting only}
@16_40c %routine romphex(%integer x)
@16_410 %routine romphex4(%integer x)
@16_414 %routine romphex2(%integer x)
@16_418 %routine romphex1(%integer x)
@16_41c %routine rompsym(%integer x)
@16_420 %routine rompstr(%string(255)s)

! Procedures defined in SUPER

%systemroutinespec return from interrupt
%systemroutinespec wait for interrupt
%systemroutinespec add interrupt handler(%integer level)
%systemroutinespec int signal semaphore(%record(semaphore fm)%name s)
%externalroutinespec become process(%integer space)        {booting only}

%systemroutinespec move to sr(%integer x)
%systemintegerfnspec or to sr(%integer x)

%systemintegerfnspec cpu time
%systemintegerfnspec elapsed time

%systemintegerfnspec priority
%systemintegerfnspec privilege
%systemintegerfnspec time slice size

%systemroutinespec set priority(%integer prio)
%systemroutinespec set privilege(%integer onoff)
%systemroutinespec set time slice size(%integer ticks)

%externalrecord(process fm)%mapspec create subprocess -
    (%integer size,pc,%string(63)s)

%systemrecord(semaphore fm)%mapspec create semaphore(%string(255)id,%integer c)
%systemrecord(semaphore fm)%mapspec lookup semaphore(%string(255)id)
%systemroutinespec delete semaphore(%record(semaphore fm)%name s)

%systemrecord(mailbox fm)%mapspec create mailbox(%string(255)id,%record(semaphore fm)%name s)
%systemrecord(mailbox fm)%mapspec lookup mailbox(%string(255)id)
%systemroutinespec delete mailbox(%record(mailbox fm)%name m)

%systemroutinespec signal semaphore(%record(semaphore fm)%name s)
%systemroutinespec semaphore wait(%record(semaphore fm)%name s)
%systemintegerfnspec semaphore counter(%record(semaphore fm)%name s)

%systemrecord(message fm)%mapspec acquire message buffer
%systemroutinespec return message buffer(%record(message fm)%name b)
%systemroutinespec put message(%record(message fm)%name m,%record(mailbox fm)%name b)
%systemrecord(message fm)%mapspec get message (%record(mailbox fm)%name mailbox)
%systemroutinespec send message(%record(message fm)%name m,%record(mailbox fm)%name b)
%systemrecord(message fm)%mapspec receive message (%record(mailbox fm)%name mailbox)

%systemroutinespec setup queue (%record(queue fm)%name queue)
%systempredicatespec queue empty(%record(queue fm)%name queue)
%systemroutinespec enqueue(%record(queue fm)%name cell,queue)
%systemroutinespec requeue(%record(queue fm)%name cell,queue)
%systemrecord(*)%mapspec dequeue(%record(queue fm)%name queue)
%systemrecord(*)%mapspec unqueue(%record(queue fm)%name queue)
%systemroutinespec exqueue(%record(queue fm)%name cell)
%systemroutinespec inqueue(%record(queue fm)%name c,l,r)

%systemintegerfnspec freestore
@(16_3f00-6*20)%integerfn heapget(%integer x)
%systemroutinespec heapput(%integer pos)
%systemstring(*)%mapspec newstring(%string(*)%name s)
%systemrecord(fcb fm)%mapspec new fcb(%string(*)%name f)

%systemroutinespec move(%integer bytes,%name from,to)

! clock

%externalroutinespec submit timer request -
  (%record(semaphore fm)%name s,%integer id,when)
%externalroutinespec submit regular timer request -
  (%record(semaphore fm)%name s,%integer id,interval)
%externalroutinespec cancel timer request(%integer id)
%externalroutinespec cancel all timer requests
%externalroutinespec wait(%integer ms)

! hex

%systemroutinespec phex1(%integer x)
%systemroutinespec phex2(%integer x)
%systemroutinespec phex4(%integer x)
%systemroutinespec phex(%integer x)
%systemintegerfnspec rhex

! names

%externalroutinespec define local logical name(%string(255)key,info)
%externalroutinespec define global logical name(%string(255)key,info)
%externalroutinespec define logical name(%string(255)key,info)
%externalroutinespec next local logical name(%string(*)%name key,info)
%externalroutinespec next global logical name(%string(*)%name key,info)
%externalpredicatespec translated local logical name(%string(*)%name s)
%externalpredicatespec translated global logical name(%string(*)%name s)
%externalpredicatespec translated logical name(%string(*)%name s)
%externalstring(255)%fnspec translate logical name(%string(255)name)

%externalroutinespec define local command symbol(%string(255)key,info)
%externalroutinespec define global command symbol(%string(255)key,info)
%externalroutinespec define command symbol(%string(255)key,info)
%externalroutinespec next local command symbol(%string(*)%name key,info)
%externalroutinespec next global command symbol(%string(*)%name key,info)
%externalpredicatespec translated local command symbol(%string(*)%name s)
%externalpredicatespec translated global command symbol(%string(*)%name s)
%externalpredicatespec translated command symbol(%string(*)%name s)
%externalstring(255)%fnspec translate command symbol(%string(255)name)

%externalroutinespec standardise filename(%string(*)%name s)

! io

%recordformat fcb fm -
  (%integer p,l,bs,bl,fs,fl,fastpc,gla,pc,mode,
   %record(fcb fm)%name next,%string(255)%name prompt,
   %integer a,b,c,d,%string(255)filename)

%constinteger -
  cbop close   = 0,
  cbop abort   = 1,
  cbop flush   = 2,
  cbop refresh = 3,
  cbop write   = 4,
  cbop read    = 5

%constinteger -
  fop logout  = 0,
  fop login   = 1,
  fop quote   = 2,
  fop pass    = 3,
  fop openi   = 4,
  fop openo   = 5,
  fop openm   = 6,
  fop opena   = 7,
  fop credir  = 8,
  fop delete  = 9,
  fop info    = 10,
  fop rename  = 11,
  fop copy    = 12,
  fop permit  = 13,
  fop stamp   = 14,
  fop time    = 15,
  fop special = 16,
  fop oldfinfo=17

